package georouting.planarization;

import georouting.Edge;
import georouting.Node;
import georouting.PlanarizationAlgorithm;
import georouting.node.ImaginaryNode;
import java.awt.geom.Point2D;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Area;
import java.util.List;
import java.util.Iterator;

/**
 * <p>
 * The gabrial graph performs planarization by removing edges in the following way:<br/>
 * <ol>
 *  <li>Let e be an element of the set of edges</li>
 *  <li>Construct 2 circles (c1 and c2). Both are have a radius equal to the distance between e.n1 and e.n2</li>
 *  <li>Position these circles so one is centered at e.n1 and the other at e.n2</li>
 *  <li>Let a be the intersection of c1 and c2.</li>
 *  <li>if the unique set constructed of N(e.n1) U N(e.n2) contains a vetex that exists in the space of a, then remove e form the set of edges</li>
 * </ol>
 * </p>
 */
public class GabrialGraph extends PlanarizationAlgorithm
{

  public List<Edge> planarize(List<Edge> edges)
  { // the return is redundent, the array is going to be modified during the algorithm
    // which will affect the call to neighborhood, thus this approch will actually work
    // (recall that neighborhood is resolved based on edge connections)
      
    Iterator<Edge> edge_it = edges.iterator();
    while(edge_it.hasNext()){
       Edge edge = edge_it.next();
       
       float midX = (edge.n1().x() - edge.n2().x()) / 2.f + edge.n2().x();
       float midY = (edge.n1().y() - edge.n2().y()) / 2.f + edge.n2().y();
       float rad  = edge.n1().distance(edge.n2()) / 2.f;

       ImaginaryNode midPt = new ImaginaryNode(midX, midY);
       
       for(Node neighbor : edge.n1().neighborhood()){
           if(neighbor != edge.n2() && neighbor.distance(midPt) <= rad){
               edge_it.remove();
               break;
           }
       }
    }      
    return edges;
  }
}
