package georouting.simulation;

import georouting.ConnectivityContract;
import georouting.Node;
import georouting.PlanarizationAlgorithm;
import georouting.TraversalAlgorithm;
import georouting.Visualizer;
import georouting.connectivity.*;
import georouting.graph.CFRGraph;
import georouting.graph.StaticGraph;
import georouting.planarization.GabrialGraph;
import georouting.simulation.DataSet;
import georouting.traversalAlgorithms.*;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Random;
import java.util.logging.Level;
import java.util.logging.Logger;

public class StaticSimCFRM {

    private static final int computations_ = 1000;
    private static final double unitDistance_ = 50.0;
    private static final ConnectivityContract cc_ = new StaticUnitDiskConnectivity(unitDistance_);
    private static final PlanarizationAlgorithm algo_ = new GabrialGraph();
    private static final PostOffice po_ = PostOffice.instance();

    // Utilities
    private static final Random rand_ = new Random();
    private static final DataSet overhead_ = new DataSet();
    private static final DataSet delivery_ = new DataSet();
    private static final DataSet latency_ = new DataSet();

    public static void main(String[] args) {
        TraversalAlgorithm message;
        StaticGraph graph;
        Node source;
        ArrayList<Node> dests = new ArrayList<>();

        PrintWriter out = null;

        try {
            FileWriter fw = new FileWriter("GMP-CFR-density-results.txt", false);
            out = new PrintWriter(fw);
        } catch (IOException ex) {
            Logger.getLogger(StaticSim.class.getName()).log(Level.SEVERE, null, ex);
        }

        Visualizer visual;

        int size = 10;

        // Print result file header
        //out.println("Targets\tOverhead\tLatency");
        out.println("Targets\tOverhead\tLatency");
        out.println("\tAverage\tSt.Dev\tAverage\tSt.Dev");
        for (int i = 0; i < 40; i++) {
            out.print("-");
        }
        out.println();

        for (double targets = 0.05; targets <= 0.05; targets += 0.05) {
            for (double density = 1; density <= 10; density++) {
                System.out.println("Density - " + density + "/20");
                //System.out.println("Targets - " + targets + "/0.50");
                boolean error = false;
                for (int comp = 0; comp < computations_; comp++) {
                    do {
                        // Setup graph
                        graph = new CFRGraph(density, size * unitDistance_, size * unitDistance_, cc_);
                        graph.planarize(algo_);
                        dests.clear();

                        source = graph.nodes().get(0);
                        ArrayList<Node> available = new ArrayList<>();

                        for (Node node : graph.nodes()) {
                            if (node.id() != 0) {
                                available.add(node);
                            }
                        }

                        int picks = (int) Math.ceil(targets * (graph.nodes().size() - 1));

                        while (picks > 0) {
                            Node pick = available.get(rand_.nextInt(available.size() - 1) + 1);
                            if (graph.isConnected(source, pick)) {
                                dests.add(pick);
                                available.remove(pick);
                            }
                            picks--;
                        }
                    } while (dests.isEmpty());

                    po_.reset();

                    //for(Node n : dests)
                    //    System.out.println(n.id());
                    GMP_dev.originTree_ = null;
                    message = po_.generate(new GMP_dev(graph, source, dests, unitDistance_, algo_));
                    //message = new CFRM_GMP(graph, source, dests, unitDistance_, algo_);
                    //message = po_.generate(new CFRM(graph, source, dests.get(0), algo_));
                    //CFRM.setCastingTargets(dests, unitDistance_, source);
                    GMP_dev.setCastingTargets(dests);
                    //((CFRM)message).setCastingTargets(dests);
                    //((CFRM)message).initiate();
                    /*System.out.print("Targets b4: ");
                    for(Node d : GMP_dev.targets())
                        System.out.print(d.id() + " ");
                    System.out.println("");
                            */
                    //System.out.println("Del: " + GMPs.delivery());

                    //System.out.println("Initial:\n" + po_);
                    //visual = new Visualizer(message);
                    //visual.begin();
                    int count = 0;
                    String diagnostic = "";
                    //((GMP_dev)message).printTree();
                    //((GMP_dev)message).printSubtree();
                    
                    while (!po_.isEmpty()) {
                        GMP_dev.increaseStep(1);
                        po_.advance();
                        diagnostic += po_ + "\n";
                        //System.out.println(po_);
                        
                        if (++count > 10000) {
                            System.out.println(diagnostic);
                            visual = new Visualizer(message);
                            visual.begin();
                            error = true;
                            break;
                        }
                    }
                                                       
                    if (error) {
                        break;
                    }
                    //System.out.println(GMPs.delivery());
                    //System.out.println(GMPs.latency());

                    if (comp % (100) == 0 && comp != 0) {
                        System.out.println("comp " + comp + "/" + computations_);
                    }

                    // Record computation results
                    overhead_.add(density, (double) po_.packets());
                    //System.out.println("Del: " + GMP_dev.delivery()); 
                    delivery_.add(density, GMP_dev.delivery());
                    if (GMP_dev.latency() != -1) {
                        latency_.add(density, GMP_dev.latency());
                    }
                    /*
                    if(GMP_dev.delivery() < 1){
                        System.out.println(diagnostic);
                        break;
                    }else{
                        visual.dispose();                           
                    }
                    */
                }
                // Log experimental results

                out.print(density);
                out.print("\t");
                out.print(overhead_.average(density));
                out.print("\t");
                out.print(overhead_.stDev(density));
                out.print("\t");
                out.print(latency_.average(density));
                out.print("\t");
                out.print(latency_.stDev(density));
                out.print("\t");
                out.print(delivery_.average(density));
                out.print("\t");
                out.print(delivery_.stDev(density));
                if(error)
                    out.print(" -E");
                out.println();
                out.flush();

                //System.out.println("overhead: " + po_.packets());
                //System.out.println("latency: " + GMP_dev.latency());
                //System.out.println(delivery_.average(density));
            }
        }

        out.close();
        System.out.println("Finished!");
    }
}
