/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package georouting.traversalAlgorithms;

import georouting.Edge;
import georouting.Graph;
import georouting.Node;
import georouting.PlanarizationAlgorithm;
import georouting.node.CFRNode;
import georouting.node.ImaginaryNode;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.event.KeyEvent;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.TreeMap;

/**
 *
 * @author Jordan
 */
public class CFRM_GMP extends CFR{
    protected ArrayList<Edge> junctures_ = new ArrayList<>();
    protected SteinerTree tree_;
    
    public CFRM_GMP(Graph g, Node s, ArrayList<Node> targets, double unitDistance, PlanarizationAlgorithm pa){
        super(g, s, targets.get(0), pa);
        seqNo_ = 0;
        id_ = 0;
        tree_ = new SteinerTree(s, targets, unitDistance);
        junctures_ = getEdges(tree_.root(), junctures_);
        initiate();
    }
    
    public CFRM_GMP(Graph g, Node s, Node d, SteinerTree tree, PlanarizationAlgorithm pa){
        super(g, s, d, pa);
        seqNo_ = 0;
        id_ = 0;
        tree_ = tree;
        junctures_ = getEdges(tree_.root(), junctures_);
        initiate();
    }
    
    public CFRM_GMP(Graph g, Node s, Node f, Node c, Node t, int id, ArrayList<Node> path, SteinerTree tree, PlanarizationAlgorithm pa){
        super(g, s, f, c, t, id, path, pa);
        tree_ = tree;
        junctures_ = getEdges(tree_.root(), junctures_);
    }
        
    @Override
    public String name(){
        return "CFRM";
    }
    
    protected void visit(Node n){
        GMP_dev.visit(n, _container);
    }
    
    @Override
    protected void advanceAlgorithm()
    {
        if(!done()){
            if(!((CFRNode)target_).hasMate(this)){                
                {                    
                    moveTo(target_);
                    visit(_current);
                    target_ = bestNeighbor();                    
                    
                    if(juncture(_previous) || juncture(target_))
                        split(_previous);            
                }
            }
            else{
                PostOffice.instance().addPacket();
                discard();
            }
        }
    }
    
    @Override
    public void initiate()
    {
        Node left, right;
              
        if(_current.neighborhood().size() > 1){
            TreeMap<Double, Node> sorted = new TreeMap<>();
            double zero = getAngle(_destination);
      
            for(Node n : _current.neighborhood()){
                double angle = getAngle(n) - zero;
                if(angle < 0) angle += 2 * Math.PI;   
                sorted.put(angle, n);
            }
              
            if(sorted.firstKey() > 0.0){
                left = sorted.lastEntry().getValue();
                right = sorted.firstEntry().getValue();
            }
            else if(_current.neighborhood().size() > 2){
                // On S-D line and multiple secondary winners to choose from
                Node sd = sorted.firstEntry().getValue();
                sorted.remove(sorted.firstKey());

                if(2 * Math.PI - sorted.lastKey() < sorted.firstKey()){
                    left = sorted.lastEntry().getValue();
                    right = sd;
                }
                else{
                    left = sd;
                    right = sorted.firstEntry().getValue();
                }
            }
            else{
                // On S-D Line and single secondary
                left  = (sorted.lastKey() > Math.PI ? sorted.lastEntry().getValue() : sorted.firstEntry().getValue());
                right = (sorted.lastKey() > Math.PI ? sorted.firstEntry().getValue() : sorted.lastEntry().getValue());
            }
        }
        else{
            left = _current.neighborhood().get(0);
            right = left;
        }        
        
        target_ = left;
        _dir = Direction.CLOCKWISE;
        CFRM_GMP mimic = new CFRM_GMP(_container, _source, _destination, _current, right, id_, path_, tree_, _pAlg);
        mimic._dir = Direction.COUNTERCLOCKWISE;
    
        PostOffice.instance().add(mimic);
        ((CFRNode)_current).receive(this);
        ((CFRNode)_current).receive(mimic);
        
        split(right);
        
        visit(_current);
    }
    
    @Override
    public void split(Node origin){
        if(_current.neighborhood().size() > 1){
            TreeMap<Double, Node> sorted = new TreeMap<>();
            double zero = getAngle(origin);
                
            for(Node n : _current.neighborhood()){
                if(!sorted.containsValue(n)){
                    Double angle = getAngle(n) - zero;
                    if(angle < 0.0) angle += 2 * Math.PI;
                    sorted.put(angle, n);
                }
            }
            if(_dir == Direction.COUNTERCLOCKWISE)
                sorted.put(10.0, sorted.remove(0.0));
        
            Iterator<Node> first  = sorted.values().iterator();
            Iterator<Node> second = sorted.values().iterator();
            second.next();
        
            CFRNode current = (CFRNode)_current;

            while(second.hasNext()){
                Node left  = first.next();
                Node right = second.next();
                
                if(juncture(left) || juncture(right)){
                    CFRM_GMP fwdLeft  = new CFRM_GMP(_container, _source, _destination, _current, left, ++seqNo_, path_, tree_, _pAlg);
                    CFRM_GMP fwdRight = new CFRM_GMP(_container, _source, _destination, _current, right, ++seqNo_, path_, tree_, _pAlg);
                    fwdRight.flipDirection();
                
                    current.receive(fwdLeft);
                    current.receive(fwdRight);
                    PostOffice.instance().add(fwdLeft);
                    PostOffice.instance().add(fwdRight);                   
                }
            }
        }
    }
    
    static protected ArrayList<Edge> getEdges(StNode n, ArrayList<Edge> edgeList){
        if(n.parentEdge() != null)
            edgeList.add(n.parentEdge());
        
        for(StNode child : n.children())
            edgeList = getEdges(child, edgeList);
        
        return edgeList;
    }
    
    public boolean juncture(Node target){        
        Edge tEdge = null;       
        try{ tEdge = new Edge(_current, target); }
        catch(Exception e) { System.err.println("Critical Error"); System.exit(1); }
        
        for(Edge edge : junctures_){
            if(tEdge.intersects(edge) || edge.intersects(target) || edge.intersects(_current) || tEdge.intersects(xd()))
                return true;
        }
        return false;
    }
     
    @Override
    public void visPaint(Graphics g)
    {
        super.visPaint(g);
                
        for(Edge edge : junctures_)
            edge.paint(g, Color.cyan);
                    
        canvas().repaint();
    }
    
    @Override
    public void processKeyReleased(KeyEvent e)
    {
        super.processKeyReleased(e);         
    }    
}

