package georouting.traversalAlgorithms;

import georouting.Graph;
import georouting.Node;
import georouting.PlanarizationAlgorithm;
import georouting.node.ImaginaryNode;
import java.util.ArrayList;
import java.util.HashSet;

public class GFGM extends GFG {

    private int id_;
    static private ArrayList<Node> targets_ = new ArrayList<>();
    protected static HashSet<Node> visited_ = new HashSet<>();
    protected static double latency_;
    protected static double steps_;
    private static int seqNo_ = 0;
    
    public GFGM(Graph g, Node s, ArrayList<Node> targets, PlanarizationAlgorithm pa) {
        super(g, s, targets.get(0), pa);
        visit(_current, _container);
        
        for(int i = 1; i < targets.size(); i++){
            GFGM messenger = new GFGM(g, s, targets.get(i), pa);
            PostOffice.instance().add(messenger);
        }
        PostOffice.instance().merge();
    }
    
    public GFGM(Graph g, Node s, Node d, PlanarizationAlgorithm pa) {
        super(g, s, d, pa);
        visit(_current, _container);
    }

    @Override
    protected void moveTo(Node n) {
        super.moveTo(n);
        PostOffice.instance().addPacket();
        visit(_current, _container);
    }
    
    static public void setCastingTargets(ArrayList<Node> targets) {
        targets_.clear();

        for (Node target : targets) {
            targets_.add(target);
        }

        visited_.clear();
        latency_ = -1;
        steps_ = 0;
    }

    static public void visit(Node n, Graph graph) {
        if (targets_.remove(n)) {
            if (targets_.isEmpty()) {
//                System.out.println("Last target:\t" + n.id());
                SHORTEST optimal = new SHORTEST(graph, graph.nodes().get(0), n);
//                SHORTEST optimal = new SHORTEST(_container, _source, n);
                while (!optimal.done()) {
                    optimal.advance();
                }
//                System.out.println("Optimal path:\t" + optimal.hops());

//                System.out.println(steps_);
                latency_ = (optimal.hops() > 0 ? steps_ / optimal.hops() : 1);

//                if(latency_ < 1){
//                    System.out.println("Last target:\t" + n.id());
//                    System.out.println("Steps:\t" + steps_ + ", optimal: " + optimal.hops());
//                    System.out.println("Latency:\t" + latency_);
//                }  
            }
            visited_.add(n);
            //System.out.println("Visited: " + n.id());
        }
    }

    public static boolean delivered() {

        if (!targets_.isEmpty()) {
            for (Node n : targets_) {
                System.out.println(Integer.toString(n.id()));
            }
        }
        return targets_.isEmpty();
    }

    public static int castSize() {
        return targets_.size();
    }

    public static int visited() {
        return visited_.size();
    }

    public static double delivery() {
        /*
         System.out.println("targets: ");
         for (Node target : targets_) {
         System.out.print(target.id() + " ");
         }
         System.out.println();
         System.out.print("visited: ");
         for (Node visited : visited_) {
         System.out.print(visited.id() + " ");
         }
         System.out.println();
         */
        return (double) visited() / (visited() + castSize());
    }

    public static void increaseStep(double amount) {
        steps_ += amount;
    }

    public static double latency() {
        return latency_;
    }

    static public ArrayList<Node> targets() {
        return targets_;
    }
}
