/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package georouting.traversalAlgorithms;

import georouting.Edge;
import georouting.Graph;
import georouting.Node;
import georouting.PlanarizationAlgorithm;
import georouting.node.CFRNode;
import georouting.node.ImaginaryNode;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.event.KeyEvent;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.TreeMap;

/**
 *
 * @author Jordan
 */
public class GMP extends FACE{
    protected static ArrayList<Node> castTargets_ = new ArrayList<>();
    protected static ArrayList<Node> visited_ = new ArrayList<>();
    
    protected ArrayList<ImaginaryNode>  destV_ = new ArrayList<>();
    protected static int                broadcasts_ = 0;
    protected static double             loss_ = 0;
    protected static double             latency_;
    protected static double             steps_;
    protected static double             unitDistance_;
    static int                          seqNo_;
    private boolean                     discarded_;
    protected SteinerNode               steinerTree_;
    private ArrayList<Edge>             treeEdges_ = new ArrayList<>();
    protected boolean                   perimeter_;
    protected int                       id_;
    
    int count_ = 0;
    
    public GMP(Graph g, Node s, PlanarizationAlgorithm pa, double unitDistance){
        super(g, s, s, pa);
        id_ = ++seqNo_;
        discarded_ = false;
        unitDistance_ = unitDistance;
        steinerTree_ = null;   
        perimeter_ = false;
    }
    
    public GMP(Graph g, Node s, ArrayList<ImaginaryNode> dests, PlanarizationAlgorithm pa, double unitDistance){
        super(g, s, s, pa);
        id_ = ++seqNo_;
        discarded_ = false;
        unitDistance_ = unitDistance;
        steinerTree_ = null;
        for(ImaginaryNode d : dests)
            destV_.add(d);
        perimeter_ = false;
    }
    
    public static void setCastingTargets(ArrayList<Node> targets){        
        castTargets_.clear();
        for(Node target : targets){
            castTargets_.add(new ImaginaryNode(target.x(), target.y()));
        }
        
        visited_.clear();
        broadcasts_ = 0;
        latency_ = -1;
        steps_ = 0;
        seqNo_ = 0;
    }
    
    @Override
    public String name(){
        return "LGS";
       
    }
    
    static public void setLatency(double latency){
        latency_ = latency;
    }
    
    protected void visit(Node n){
        destV_.remove(n);
        if(castTargets_.remove(n)){
            if(castTargets_.isEmpty()){
//                System.out.println("Last target:\t" + n.id());
                SHORTEST optimal = new SHORTEST(_container, _container.nodes().get(0), n);
//                SHORTEST optimal = new SHORTEST(_container, _source, n);
                while(!optimal.done()) optimal.advance();
//                System.out.println("Optimal path:\t" + optimal.hops());

//                System.out.println(steps_);
                latency_ = (optimal.hops() > 0 ? steps_ / optimal.hops() : 1);

//                if(latency_ < 1){
//                    System.out.println("Last target:\t" + n.id());
//                    System.out.println("Steps:\t" + steps_ + ", optimal: " + optimal.hops());
//                    System.out.println("Latency:\t" + latency_);
//                }
                
            }
        }
        if(!visited_.contains(n))
            visited_.add(n);
    }
    
    @Override
    public void advanceAlgorithm()
    {
        if(!done()){
            visit(_current);
                    
            if(!destV_.isEmpty()){
                rrSTR(destV_);
                
                int destCount = subTreeCount(steinerTree_) - 1;
                //System.out.println("Tree count: " + destCount);

                ArrayList<SteinerNode> pivots = new ArrayList<>();
                for(SteinerNode stn : steinerTree_.children){
                    pivots.add(stn);
                }

                int pivot_i = 0;
                
                while(pivot_i < pivots.size()){                
                    SteinerNode pivot = pivots.get(pivot_i);
                    Node next = null;

                    for(Node n : _current.neighborhood()){
                        if(distanceSum(n, pivot) < distanceSum(_current, pivot)){
                            if(next == null)
                                next = n;
                            else if(n.distance(pivot.destination) < next.distance(pivot.destination))
                                next = n;
                        }
                    }

                    if(next != null){
                        pivots.remove(pivot);                    
                        ArrayList<ImaginaryNode> vp = new ArrayList<>();
                        //System.out.println("Copy for target " + next.id() + " with tree:");
                        subDestinations(pivot, vp, 0);
                        
                        if(!pivots.isEmpty()){
/*                            System.out.print("message " + id_ + " split to " + next.id() + " with dests: ");
                            for(Node v : vp)
                                System.out.print(v.id() + " ");
                            System.out.println();
*/                        }
                        if(perimeter_ && vp.size() == destCount){
                            pivots.clear();
                            for(ImaginaryNode v : vp)
                                pivots.add(new SteinerNode(v));
                        }
                        else{
                            perimeter_ = false;

                            GMP copy = new GMP(_container, _source, vp, _pAlg, unitDistance_);
                            copy._current = _current;

                            PostOffice.instance().add(copy);
                            copy.moveTo(next);
                        }
                    }
                    else if(!pivot.children.isEmpty()){
                        SteinerNode last = pivot.children.get(pivot.children.size() - 1);
                        pivots.add(last);
                        pivot.children.remove(last);
                        steinerTree_.children.add(last);

//                        System.out.println("Appropriate next not found; pruning branch " + last.destination.id());                        

                        if(pivot.children.size() == 1 && pivot.virtual){
                            SteinerNode o = pivot.children.get(0);

//                            System.out.println("Pruned half of a virtual split with dest " + o.destination.id() + "; cutting out the middle man.");

                            pivot.children.remove(o);
                            steinerTree_.children.add(o);
                            pivots.add(o);
                            pivots.remove(pivot);
                        }
                        //Else: cont with same p
                    }
                    else{
//                        System.out.println("leaf pivot " + pivot.destination.id() + " hit local max at " + _current.id());

                        if(pivot.virtual)
                            pivots.remove(pivot);
                        else
                            pivot_i++;
                    }
                }
                
                if(!pivots.isEmpty()){
                    // PERIMETER MODE
                    if(pivots.size() == 1){
                        // If only one pivot left, GFG it.
                        SteinerNode p = pivots.get(0);                        
//                        System.out.println("Destination " + p.destination.id() + " hit a local maximum at " + _current.id() +
//                               "; GFG routing.");
                        GFG_GMP sprinter = new GFG_GMP(_container, _current, p.destination, _pAlg);
                        PostOffice.instance().add(sprinter);
                    }                                        
                    // Try to PERI mode all problem pivots, according to the centroid of their destinations.
                    else{
                        ArrayList<ImaginaryNode> vp = new ArrayList<>();
                        double x_bar = 0, y_bar = 0;

//                        System.out.print("Pivots: ");
                        for(SteinerNode p : pivots){
                            if(!p.virtual){
//                                System.out.print(p.destination.id() + " ");
                                vp.add(p.destination);
                                x_bar += p.destination.x();
                                y_bar += p.destination.y();
                            }
                        }
 //                       System.out.println(" hit local maximum at " + _current.id());

                        x_bar /= vp.size();
                        y_bar /= vp.size();
                            
                        if(vp.size() < destCount || _previous == null || !perimeter_){
                            // First time in PERI-mode for this sub-group
 //                           System.out.println("Message " + id_ + " initiating PERI-route ");

                            GMP copy = new GMP(_container, _current, vp, _pAlg, unitDistance_);                           
                            copy._destination = new ImaginaryNode(x_bar, y_bar);
                            copy.perimeter_ = true;

                            PostOffice.instance().add(copy);
                            copy.firstRun();
                        }
                        else{
                            // Continued PERI-route for this sub-group
//                            System.out.println("Message " + id_ + " continuing PERI-route with prev: " + _previous.id());

                            GMP copy = new GMP(_container, _source, vp, _pAlg, unitDistance_);
                            copy._current = _current;
                            copy._dir = _dir;
                            copy._previous = _previous;
                            copy._destination = _destination;
                            copy.perimeter_ = true;

                            PostOffice.instance().add(copy);
                            copy.moveTo(bestNeighbor());     
                        }
                    }
                }            
            }
            
            discarded_ = true;
        }
    }
               
    protected void subDestinations(SteinerNode p, ArrayList<ImaginaryNode> dests, int t){
        if(!p.virtual)
            dests.add(p.destination);
        //for(int tabs = 0; tabs < t; tabs++)
        //    System.out.print("\t");
        //System.out.println(p.destination.id());
        for(SteinerNode child : p.children)
            subDestinations(child, dests, t + 1);
    }
    
    protected double distanceSum(Node n, SteinerNode p){
        double distance = p.virtual ? 0.0 : n.distance(p.destination);
        //System.out.println("d(" + n.id() + ", " + p.destination.id() + ") = " + distance);
        
        for(SteinerNode child : p.children)
            distance += distanceSum(n, child);
        
        return distance;
    }
    
    protected int subTreeCount(SteinerNode p){
        int total = 0;
        if(!p.virtual) total++;
        
        for(SteinerNode child : p.children)
            total += subTreeCount(child);
        
        return total;
    }
            
    protected void moveTo(Node n){
        _hops ++;
        _previous = _current;
        _current = n;
        PostOffice.instance().addPacket();
    }
    
    protected void rrSTR(ArrayList<ImaginaryNode> destV){
        ArrayList<SteinerNode> dests = new ArrayList<>();
        for(ImaginaryNode d : destV){
            if(!visited_.contains(d))
                dests.add(new SteinerNode(d));            
        }

        ArrayList<GMPTuple> destPairs = new ArrayList<>();
        for(SteinerNode u : dests){
            for(SteinerNode v : dests){
                if(u.destination.id() <= v.destination.id())
                    destPairs.add(new GMPTuple(u, v));
            }
        }
        
        try{
            destPairs.sort(null);
        }
        catch(Exception e){
            System.out.println("---------------------------------------------");
            System.out.println("- Destination Pairs");
            System.out.println("---------------------------------------------");
            for(GMPTuple tup : destPairs)
                System.out.println(tup);
            System.out.println(e);            
        }
        
        /*
        for(GMPTuple tup : destPairs){
            System.out.println(tup.u.destination.id() + ", " + tup.v.destination.id() + ", " + tup.t.destination.id() + "\t: " +
                    tup.reductionRatio_);
        }
        */
               
        steinerTree_ = new SteinerNode(new ImaginaryNode(_current.x(), _current.y()));
        SteinerNode s = steinerTree_;
        boolean updateSort = false;
        boolean updatePairs = false;
        
        while(!destPairs.isEmpty()){
            GMPTuple bestPair = destPairs.get(0);
            SteinerNode u = bestPair.u;
            SteinerNode v = bestPair.v;
            SteinerNode t = bestPair.t;
                        
            if(u == v){
                //System.out.println("u == v : " + u.destination.id() + ", " + v.destination.id());
                s.addChild(u);
                u.active = false;
                updatePairs = true;
            }            
            else if(t.collocated(s)){
                //System.out.println("t collocated s : " + u.destination.id() + ", " + v.destination.id());
                s.addChild(u);
                s.addChild(v);
                u.active = false;
                v.active = false;
                updatePairs = true;

            }            
            else if(t.collocated(u)){
                //System.out.println("t collocated u : " + u.destination.id() + ", " + v.destination.id());
                u.addChild(v);
                v.active = false;
                updatePairs = true;
            }
            else if(t.collocated(v)){
                //System.out.println("t collocated v : " + u.destination.id() + ", " + v.destination.id());
                v.addChild(u);
                u.active = false;
                updatePairs = true;
            }
            
            else if(s.distance(u) < unitDistance_ && s.distance(v) < unitDistance_){
                //System.out.println("d(s,u) and d(s,v) < RR : " + u.destination.id() + ", " + v.destination.id());
                destPairs.remove(bestPair);
            }
            
            else if(s.distance(u) < unitDistance_){
                if(unitDistance_ + t.distance(u) + t.distance(v) > s.distance(u) + s.distance(v)){
                    //System.out.println("d(s,u) < RR AND RR + d(t,u) + d(t,v) > d(s,u) + d(s,v) : " + u.destination.id() + ", " + v.destination.id());                    
                    destPairs.remove(bestPair);
                }
                else{
                    //System.out.println("d(s,u) < RR BUT RR + d(t,u) + d(t,v) <= d(s,u) + d(s,v) : " + u.destination.id() + ", " + v.destination.id());
                    u.addChild(v);
                    v.active = false;
                    updatePairs = true;
                }                   
            }            
            else if(s.distance(v) < unitDistance_){
                if(unitDistance_ + t.distance(u) + t.distance(v) > s.distance(u) + s.distance(v)){
                    //System.out.println("d(s,v) < RR AND RR + d(t,u) + d(t,v) > d(s,u) + d(s,v) : " + u.destination.id() + ", " + v.destination.id());
                    destPairs.remove(bestPair);
                }
                else{
                    //System.out.println("d(s,u) < RR BUT RR + d(t,u) + d(t,v) <= d(s,u) + d(s,v) : " + u.destination.id() + ", " + v.destination.id());
                    v.addChild(u);
                    u.active = false;
                    updatePairs = true;
                }                   
            }
            
            else if(s.distance(t) < unitDistance_ &&
                    (unitDistance_ + t.distance(u) + t.distance(v) > s.distance(u) + s.distance(v))){
                //System.out.println("d(s,t) < RR AND RR + d(t,u) + d(t,v) > d(s,u) + d(s,v) : " + u.destination.id() + ", " + v.destination.id());
                s.addChild(u);
                s.addChild(v);
                u.active = false;
                v.active = false;
                updatePairs = true;
            }            
            else{
                // New Virtual Destination
                //System.out.println("new virtual : " + u.destination.id() + ", " + v.destination.id());
                
                t.addChild(u);
                t.addChild(v);
                dests.add(t);
                for(SteinerNode i : dests){
                    //if(!i.active)
                        //System.out.println("RePOOORT!");
                    destPairs.add(new GMPTuple(t, i));
                }
                
                u.active = false;
                v.active = false;
                updatePairs = true;
                updateSort = true;
            }            

            if(updatePairs){
                Iterator<SteinerNode> destIt = dests.iterator();
                Iterator<GMPTuple> pairIt = destPairs.iterator();
                       
                while(destIt.hasNext()){
                    SteinerNode dest = destIt.next();
                    if(!dest.active)
                        destIt.remove();
                }
            
                while(pairIt.hasNext()){
                    GMPTuple tup = pairIt.next();
                    if(!tup.u.active || !tup.v.active)
                        pairIt.remove();
                }
                updatePairs = false;
            }
            if(updateSort){
                destPairs.sort(null);
                updateSort = false;
            }
        }        
    }
    
    public void initiate()
    {
        destV_.clear();
        if(_current.neighborhood().size() > 0){
            for(Node d : castTargets_)
                destV_.add(new ImaginaryNode(d.x(), d.y())); 

        }
        else
            discarded_ = true;        
    }
    
    protected class GMPTuple implements Comparable{
        public double reductionRatio_ = 0;
        public SteinerNode u;
        public SteinerNode v;
        public SteinerNode t;
        
        GMPTuple(SteinerNode left, SteinerNode right){
            u = left;
            v = right;
            t = SteinerNode.steinerPoint(_current.x(), _current.y(), left.destination.x(), left.destination.x(),
                    right.destination.x(), right.destination.y());
            reductionRatio_ = t.reductionRatio(new ImaginaryNode(_current.x(), _current.y()), left.destination, right.destination);                                 
        }
       
        @Override
        public int compareTo(Object other) {
            //System.out.println("comparing: " + ((GMPTuple)other).reductionRatio_ + " to " + reductionRatio_);
            double comparison =  10000 * (((GMPTuple)other).reductionRatio_ - reductionRatio_);
                         
            if(comparison != comparison){
                System.out.println("me:" + reductionRatio_ + " them:" + ((GMPTuple)other).reductionRatio_ + " = " + comparison);
            }
            return (int)comparison;
        }
        
        @Override
        public String toString(){
            return reductionRatio_ + " - " + "u:" + u.destination.id() + ", v:" + v.destination.id() + ", s:"
                + _current.id() + ", t:" + t.destination.id();
        }
    }       
    
    @Override
    public boolean done()
    {
        return discarded_;
    }

    public static void increaseStep(double amount){        
        steps_ += amount;
//        System.out.println("steps: " + steps_);
    }
    
    public static double latency(){
        return latency_;
    }
    public static void setLoss(double loss){
        loss_ = loss;    
    }
                         
    @Override
    public void visPaint(Graphics g)
    {        
        _container.visPaint(g);                
        _source.paint(g,Color.red);
        _destination.paint(g,Color.red);
        
        for(Node t : castTargets_)
            t.paint(g, Color.yellow);
        
        //if(steinerTree_ != null)
        //    steinerTree_.paint(g);
                    
        canvas().repaint();
    }     
    
    @Override
    public void processKeyReleased(KeyEvent e)
    {
        super.processKeyReleased(e);         
    }
    
    public static void addTarget(Node n){
        castTargets_.add(n);
    }
    
    public static boolean delivered(){
        
        if(!castTargets_.isEmpty()){
            for(Node n : castTargets_){
                System.out.println(Integer.toString(n.id()));                
            }            
        }
        return castTargets_.isEmpty();
    }
    
    public static int castSize(){
//        for(Node n : castTargets_){
//                System.out.println(Integer.toString(n.id()));                
//        }
        return castTargets_.size();
    }
        
    public static int visited(){
        return visited_.size();
    }
    
    public static int broadcasts(){
        return broadcasts_;
    }
    
    @Override
    public String toString(){
        String statusStr = "{id: " + Integer.toString(id_) + ", s: " + Integer.toString(_source.id()) + ", c: " + Integer.toString(_current.id()) + ", d:";
        
        for(Node d : destV_)
            statusStr += " " + Integer.toString(d.id());
        
        statusStr += (_dir == Direction.CLOCKWISE ? ", C " : ", CC");
        statusStr += (perimeter_ ? " -PERI" : "");
        
        statusStr += (discarded_ ? "} -D" : "}");
        
        return statusStr;
    }
}

