/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package georouting.traversalAlgorithms;

import georouting.Graph;
import georouting.Node;
import georouting.PlanarizationAlgorithm;
import georouting.node.ImaginaryNode;
import java.awt.Color;
import java.awt.Graphics;
import java.util.ArrayList;
import java.util.HashSet;

/**
 *
 * @author Jordan
 */
public class GMP_backup extends PlanarTraversalAlgorithm {

    private SteinerTree tree_;
    private boolean done_;
    private int id_;
    private int targetCount_;
    private FACE guide_;

    static private ArrayList<Node> targets_ = new ArrayList<>();
    protected static HashSet<Node> visited_ = new HashSet<>();
    protected static double latency_;
    protected static double steps_;
    private static int seqNo_ = 0;

    public GMP_backup(Graph g, Node s, ArrayList<Node> targets, double unitDistance, PlanarizationAlgorithm pa) {
        super(g, s, s, pa);
        done_ = _current.neighborhood().size() <= 0 || targets.isEmpty();
        tree_ = new SteinerTree(s, targets, unitDistance);
        targetCount_ = targets.size();
        id_ = ++seqNo_;
    }

    public GMP_backup(Graph g, Node s, ArrayList<Node> targets, double unitDistance, PlanarizationAlgorithm pa, FACE guide) {
        super(g, s, s, pa);
        done_ = _current.neighborhood().size() <= 0 || targets.isEmpty();
        tree_ = new SteinerTree(s, targets, unitDistance);
        targetCount_ = targets.size();
        id_ = ++seqNo_;
        guide_ = guide;
    }
    
    public GMP_backup(Graph g, Node s, ArrayList<Node> targets, double unitDistance, PlanarizationAlgorithm pa, StNode root) {
        super(g, s, s, pa);
        // non-virtual sub-targets
        done_ = _current.neighborhood().size() <= 0 || targets.isEmpty();
        tree_ = new SteinerTree(root, unitDistance);
        targetCount_ = targets.size();
        id_ = ++seqNo_;
    }

    public GMP_backup(Graph g, Node s, ArrayList<Node> targets, double unitDistance, PlanarizationAlgorithm pa, FACE guide, StNode root) {
        super(g, s, s, pa);
        done_ = _current.neighborhood().size() <= 0 || targets.isEmpty();
        tree_ = new SteinerTree(root, unitDistance);
        targetCount_ = targets.size();
        id_ = ++seqNo_;
        guide_ = guide;
    }

    public void advanceAlgorithm() {
        if (!done()) {
            // Remove current location from remaining multicast targets
            visit(_current, _container);

            // Count of all non-virtual sub-targets
            int destCount = tree_.root().subTreeCount() - 1;

            // Copy root node's children list as pivots
            ArrayList<StNode> pivots = new ArrayList<>();
            for (StNode child : tree_.root().children()) {
                pivots.add(child);
            }

            // Set pivot index to the start (0) - using index to avoid awful
            // Java-based concurrent mod errors
            int pivot_i = 0;
            
            ArrayList<SpawnRequest> spawns = new ArrayList<>();

            // Begin main pivot loop
            while (pivot_i < pivots.size()) {
                StNode pivot = pivots.get(pivot_i);

                // This destination has been successfully reached
                if (pivot.destination() == _current) {
                    // Re-attach its children to root
                    for (int child_i = 0; child_i < pivot.children().size(); child_i++) {
                        pivots.add(pivot.children().get(child_i));
                        tree_.root().addChild(pivot.children().get(child_i));
                        pivot.children().remove(child_i);
                    }
                    // Remove this pivot as delivered
                    pivots.remove(pivot);
                } // Proceed to find a greedy neighbor
                else {
                    Node next = null;

                    // Find node with minimum distance (distance to all targets) to pivot
                    // such that said distance sum is less than current to pivot (Greedy).
                    for (Node n : _current.neighborhood()) {
                        if (pivot.distanceSum(n) < pivot.distanceSum(_current)) {
                            if (next == null || n.distance(pivot.destination()) < next.distance(pivot.destination())) {
                                next = n;
                            }
                        }
                    }
                    // If we find a greedy neighbor to go to ...
                    if (next != null) {
                        // If this split-off encompasses the entire target list, and is attempting to move that bundled
                        // list to the same node that a FACE guide just recently routed away from, skip this pivot
                        // (effectively forwarding the whole message along the FACE guide).
                        if (pivot.subTreeCount() == targetCount_ && guide_ != null && !guide_.progress(next)){
                            pivot_i++;
                        }
                        // Else, split this message to the greedy neighbor.
                        else {                
                            pivots.remove(pivot);

                            ArrayList<Node> vp = new ArrayList<>();
                            vp = pivot.subDestinations(vp, 0);
                            
                            //SpawnRequest spawn = new SpawnRequest(pivot, next, vp);
                            // Split-off message with this pivot's sub-targets                            
                            //spawns.add(spawn);
                            GMP_backup spawn = new GMP_backup(_container, next, vp, tree_.RR(), _pAlg);
                            spawn._current = _current;

                            PostOffice.instance().add(spawn);
                            spawn.moveTo(next); 
                        }
                    } // If we fail to find a greedy neighbor, and our pivot is not a leaf
                    else if (!pivot.children().isEmpty()) {
                        // Re-attach current pivot's last child to root's children.
                        StNode last = pivot.children().get(pivot.children().size() - 1);

                        pivots.add(last);
                        pivot.children().remove(last);
                        tree_.root().addChild(last);

                        // If detaching child from a virtual splitter pivot, a single unnecessary
                        // virtual -> non-virtual chain exists. Cut out the virtual middle-man.
                        if (pivot.children().size() == 1 && pivot.virtual()) {
                            StNode onlyChild = pivot.children().get(0);

                            pivot.children().remove(onlyChild);
                            tree_.root().addChild(onlyChild);
                            pivots.add(onlyChild);
                            pivots.remove(pivot);
                        }
                    } // Reached a local max for a leaf pivot
                    else {
                        // Remove it if it is a lone-virtual pivot
                        if (pivot.virtual()) {
                            pivots.remove(pivot);
                        } // Skip it for now otherwise; will be later handled in perimeter mode.
                        else {
                            pivot_i++;
                        }
                    }
                }
            }
            // End main pivot loop
            
            // Create all messages spawned from subtrees
            for(SpawnRequest req : spawns){                
                GMP_backup spawn = new GMP_backup(_container, req.next, req.targets, tree_.RR(), _pAlg);
                spawn._current = _current;
                PostOffice.instance().add(spawn);
                spawn.moveTo(req.next);                
            }            

            // If any pivots remain, they are local maxima, and perimeter routing is required.
            if (!pivots.isEmpty()) {
                // Flatten the remaining pivots into a single list of total non-virtual destinations
                ArrayList<Node> vp = new ArrayList<>();
                for (StNode pivot : pivots) {
                    vp = pivot.subDestinations(vp, 0);
                }
                /*
                 System.out.println("pivots reached local max: ");
                 for(StNode pivot : pivots)
                 System.out.print(pivot.destination().id() + " ");
                 System.out.println();
                 */
                // If only one problem target exists, GFG route it.
                //if (vp.size() == 1) {
                //    GFG_GMP sprinter = new GFG_GMP(_container, _current, vp.get(0), _pAlg);
                //    PostOffice.instance().add(sprinter);
                //} // More than one failed target exists: bundle face route around the centroid of their
                // invidual destinations.
                //else {                                        
                    CFRM_GMP sprinter = new CFRM_GMP(_container, _current, vp, tree_.RR(), _pAlg);
                    PostOffice.instance().add(sprinter);                    
                //}
            }
            done_ = true;
        }
    }

    public boolean done() {
        return done_;
    }

    protected void moveTo(Node n) {
        _current = n;
        _hops++;
        PostOffice.instance().addPacket();
    }

    static public void setCastingTargets(ArrayList<Node> targets) {
        targets_.clear();

        for (Node target : targets) {
            targets_.add(target);
        }

        visited_.clear();
        latency_ = -1;
        steps_ = 0;
    }

    static public void visit(Node n, Graph graph) {
        if (targets_.remove(n)) {
            if (targets_.isEmpty()) {
//                System.out.println("Last target:\t" + n.id());
                SHORTEST optimal = new SHORTEST(graph, graph.nodes().get(0), n);
//                SHORTEST optimal = new SHORTEST(_container, _source, n);
                while (!optimal.done()) {
                    optimal.advance();
                }
//                System.out.println("Optimal path:\t" + optimal.hops());

//                System.out.println(steps_);
                latency_ = (optimal.hops() > 0 ? steps_ / optimal.hops() : 1);

//                if(latency_ < 1){
//                    System.out.println("Last target:\t" + n.id());
//                    System.out.println("Steps:\t" + steps_ + ", optimal: " + optimal.hops());
//                    System.out.println("Latency:\t" + latency_);
//                }  
            }
            visited_.add(n);
            //System.out.println("Visited: " + n.id());
        }
    }

    public static boolean delivered() {

        if (!targets_.isEmpty()) {
            for (Node n : targets_) {
                System.out.println(Integer.toString(n.id()));
            }
        }
        return targets_.isEmpty();
    }

    public static int castSize() {
        return targets_.size();
    }

    public static int visited() {
        return visited_.size();
    }

    public static double delivery() {
        /*
         System.out.println("targets: ");
         for (Node target : targets_) {
         System.out.print(target.id() + " ");
         }
         System.out.println();
         System.out.print("visited: ");
         for (Node visited : visited_) {
         System.out.print(visited.id() + " ");
         }
         System.out.println();
         */
        return (double) visited() / (visited() + castSize());
    }

    public static void increaseStep(double amount) {
        steps_ += amount;
    }

    public static double latency() {
        return latency_;
    }

    static public ArrayList<Node> targets() {
        return targets_;
    }

    @Override
    public void visPaint(Graphics g) {
        super.visPaint(g);

        if (tree_ != null) {
            tree_.paint(g);
        }

        canvas().repaint();
    }

    @Override
    public String toString() {
        String statusStr = "{id: " + Integer.toString(id_) + ", s: " + Integer.toString(_source.id()) + ", c: " + Integer.toString(_current.id()) + ", d:";

        ArrayList<Node> dests = new ArrayList<>();
        dests = tree_.root().subDestinations(dests, 0);
        dests.remove(0);

        for (Node d : dests) {
            statusStr += " " + Integer.toString(d.id());
        }

        {
            statusStr += (done_ ? "} -D" : "}");
        }

        return statusStr;
    }
    
    public void printTree(){
        tree_.root().print(0);
    }
    
    public void printSubtree(){
        SteinerTree clone = new SteinerTree(tree_.root(), tree_.RR());
        System.out.println("----------------------------------------------------------------------------");
        System.out.println("- Cloned");
        System.out.println("----------------------------------------------------------------------------");
        clone.root().print(0);
    }
}
