/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package georouting.traversalAlgorithms;

import georouting.Edge;
import georouting.Graph;
import georouting.Node;
import georouting.PlanarizationAlgorithm;
import georouting.node.CFRNode;
import georouting.node.ImaginaryNode;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.event.KeyEvent;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.TreeMap;

/**
 *
 * @author Jordan
 */
public class GMPs extends GFG{
    protected static ArrayList<Node> castTargets_ = new ArrayList<>();
    protected static ArrayList<Node> visited_ = new ArrayList<>();
    protected static int             broadcasts_ = 0;
    protected static double          loss_ = 0;
    protected static double          latency_;
    protected static double          steps_;
    static int                       seqNo_;
    private boolean                  discarded_;
    protected StNode                 destinationTree_;
    static private ArrayList<Edge>   treeEdges_ = new ArrayList<>();
    static private SteinerTree       tree_;
       
    public GMPs(Graph g, Node s, ArrayList<Node> targets, double unitDistance, PlanarizationAlgorithm pa){
        super(g, s, s, pa);
        seqNo_ = 0;
        discarded_ = false;
        tree_ = new SteinerTree(s, targets, unitDistance);        
        destinationTree_ = tree_.root();
    }
    
    public GMPs(Graph g, Node s, StNode tree, PlanarizationAlgorithm pa){
        super(g, s, tree.destination(), pa);
        seqNo_ = 0;
        discarded_ = false;
        destinationTree_ = tree;
    }
    
    public void setCastingTargets(ArrayList<Node> targets){
        castTargets_.clear();
        for(Node target : targets)
            castTargets_.add(target);
        treeEdges_.clear();
        visited_.clear();
        broadcasts_ = 0;
        latency_ = -1;
        steps_ = 0;
    }
    
    public void initiate(){
        treeEdges_ = getEdges(destinationTree_, treeEdges_);

        visit(_current);
        split();
        PostOffice.instance().merge();
    }
    
    static protected ArrayList<Edge> getEdges(StNode n, ArrayList<Edge> edgeList){
        if(n.parentEdge() != null)
            edgeList.add(n.parentEdge());
        
        for(StNode child : n.children())
            edgeList = getEdges(child, edgeList);
        
        return edgeList;
    }
                
    @Override
    public String name(){
        return "GPS-s";
    }
    
    protected void visit(Node n){
        if(castTargets_.remove(n)){
            if(castTargets_.isEmpty()){
//                System.out.println("Last target:\t" + n.id());
                SHORTEST optimal = new SHORTEST(_container, _container.nodes().get(0), n);
//                SHORTEST optimal = new SHORTEST(_container, _source, n);
                while(!optimal.done()) optimal.advance();
//                System.out.println("Optimal path:\t" + optimal.hops());

//                System.out.println(steps_);
                latency_ = (optimal.hops() > 0 ? steps_ / optimal.hops() : 1);

//                if(latency_ < 1){
//                    System.out.println("Last target:\t" + n.id());
//                    System.out.println("Steps:\t" + steps_ + ", optimal: " + optimal.hops());
//                    System.out.println("Latency:\t" + latency_);
//                }
                
            }
        }
        if(!visited_.contains(n))
            visited_.add(n);
    }
    
    @Override
    public void advanceAlgorithm()
    {
        if(!done()){
            super.advanceAlgorithm();            
            {                  
                boolean juncture = false;

                if(destinationTree_.virtual()){
                    Edge traversed = justTraversed();
                    for(StNode child : destinationTree_.children()){
                        if(bypasses(traversed, child.parentEdge()) || _current == child.destination()){
                            juncture = true;
                            break;
                        }
                    }
                }

                if(_current == _destination || juncture){
                    split();
                }
            }
            visit(_current);
        }
    }
    
    protected boolean bypasses(Edge e, Edge line){
        double slope = (line.n1().y() - line.n2().y()) / (line.n1().x() - line.n2().x());
        double dist1 = e.n1().x() * slope - e.n1().y();
        double dist2 = e.n2().x() * slope - e.n2().y();
        
        return dist1 < 0 == dist2 < 0;
    }
    
    protected void moveTo(Node n){
        _hops ++;
        _previous = _current;
        _current = n;
        
        PostOffice.instance().addPacket();
    }    
    
    public void split(){
        discarded_ = true;
        for(StNode d : destinationTree_.children()){
            GMPs child = new GMPs(_container, _current, d, _pAlg);
            PostOffice.instance().add(child);
        }
    }
    
    @Override
    public boolean done()
    {
        return discarded_;
    }

    public static void increaseStep(double amount){        
        steps_ += amount;
//        System.out.println("steps: " + steps_);
    }
    
    public static double latency(){
        return latency_;
    }
    public static void setLoss(double loss){
        loss_ = loss;
    }    
     
    @Override
    public void visPaint(Graphics g)
    {
        _container.visPaint(g);
        _source.paint(g,Color.blue);
        _destination.paint(g,Color.red);
        
        for(Node t : castTargets_)
            t.paint(g, Color.green);
        
        for(Edge edge : treeEdges_)
            edge.paint(g, Color.cyan);
        
        destinationTree_.paint(g);
                    
        canvas().repaint();
    }
    
    @Override
    public void processKeyReleased(KeyEvent e)
    {
        super.processKeyReleased(e);         
    }
    
    public static void addTarget(Node n){
        castTargets_.add(n);
    }
    
    public static boolean delivered(){
        
        if(!castTargets_.isEmpty()){
            for(Node n : castTargets_){
                System.out.println(Integer.toString(n.id()));                
            }            
        }
        return castTargets_.isEmpty();
    }
    
    public static double delivery(){           
        return (double)visited_.size() / (visited_.size() + castTargets_.size());
    }
    
    public static int castSize(){
        
        return castTargets_.size();
    }
        
    public static int visited(){
        return visited_.size();
    }
    
    public static int broadcasts(){
        return broadcasts_;
    }
    @Override
    public String toString(){
        return super.toString();
  }
}

