/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package georouting.traversalAlgorithms;

import georouting.Edge;
import georouting.Graph;
import georouting.Node;
import georouting.PlanarizationAlgorithm;
import georouting.node.CFRNode;
import georouting.node.ImaginaryNode;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.event.KeyEvent;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.TreeMap;

/**
 *
 * @author Jordan
 */
public class LGS extends GFG{
    protected static ArrayList<Node> castTargets_ = new ArrayList<>();
    protected static ArrayList<Node> visited_ = new ArrayList<>();
    protected static int             broadcasts_ = 0;
    protected static double          loss_ = 0;
    protected static double          latency_;
    protected static double          steps_;
    static int                       seqNo_;
    private boolean                  discarded_;
    protected DestTreeNode           destinationTree_;
    static private ArrayList<Edge>   treeEdges_ = new ArrayList<>();
    
    public LGS(Graph g, Node s, PlanarizationAlgorithm pa){
        super(g, s, s, pa);
        seqNo_ = 0;
        discarded_ = false;        
        destinationTree_ = new DestTreeNode(s);        
    }
    
    public LGS(Graph g, Node s, DestTreeNode tree, PlanarizationAlgorithm pa){
        super(g, s, tree.destination, pa);
        seqNo_ = 0;
        discarded_ = false;
        destinationTree_ = tree;
    }
    
    public void setCastingTargets(ArrayList<Node> targets){
        castTargets_.clear();
        for(Node target : targets)
            castTargets_.add(target);
        treeEdges_.clear();
        visited_.clear();
        broadcasts_ = 0;
        latency_ = -1;
        steps_ = 0;
    }
    
    @Override
    public String name(){
        return "LGS";
    }
    
    protected void visit(Node n){
        if(castTargets_.remove(n)){
            if(castTargets_.isEmpty()){
//                System.out.println("Last target:\t" + n.id());
                SHORTEST optimal = new SHORTEST(_container, _container.nodes().get(0), n);
//                SHORTEST optimal = new SHORTEST(_container, _source, n);
                while(!optimal.done()) optimal.advance();
//                System.out.println("Optimal path:\t" + optimal.hops());

//                System.out.println(steps_);
                latency_ = (optimal.hops() > 0 ? steps_ / optimal.hops() : 1);

//                if(latency_ < 1){
//                    System.out.println("Last target:\t" + n.id());
//                    System.out.println("Steps:\t" + steps_ + ", optimal: " + optimal.hops());
//                    System.out.println("Latency:\t" + latency_);
//                }
                
            }
        }
        if(!visited_.contains(n))
            visited_.add(n);
    }
    
    @Override
    public void advanceAlgorithm()
    {
        if(!done()){
            super.advanceAlgorithm();            
            {                  
                if(_current == _destination){
                    visit(_current);
                    split();
                }
            }            
        }
    }
    
    protected void moveTo(Node n){
        _hops ++;
        _previous = _current;
        _current = n;
        
        PostOffice.instance().addPacket();
    }
    
    public void initiate()
    {
        if(_current.neighborhood().size() > 0){
            // calculate tree
            ArrayList<DestTreeNode> destTree = new ArrayList<>();            

            destTree.add(destinationTree_);
        
            ArrayList<Node> available = new ArrayList<>();
            for(Node d : castTargets_)
                available.add(d);

            while(!available.isEmpty()){
                float min = Float.MAX_VALUE;
                Node            child = null;
                DestTreeNode    parent = null;
                for(DestTreeNode right : destTree){
                    for(Node left : available){                    
                        if(left.distance(right.destination) < min){
                            min = left.distance(right.destination);
                            child = left;
                            parent = right;
                        }
                    }
                }
                DestTreeNode childNode = new DestTreeNode(child);
                parent.children.add(childNode);
                destTree.add(childNode);
                available.remove(child);
                
                //System.out.println("Added: " + child.id() + " to " + parent.destination.id());
                
                Edge tEdge = null;       
                try{ tEdge = new Edge(parent.destination, child); }
                catch(Exception e) { System.err.println("Critical Error"); System.exit(1); }
                treeEdges_.add(tEdge);
            }
        
        }
        
        visit(_current);
        split();
        PostOffice.instance().merge();
    }
    
    public void split(){
        discarded_ = true;
        for(DestTreeNode d : destinationTree_.children){
            LGS child = new LGS(_container, _current, d, _pAlg);
            PostOffice.instance().add(child);
        }
    }
    
    @Override
    public boolean done()
    {
        return discarded_;
    }

    public static void increaseStep(double amount){        
        steps_ += amount;
//        System.out.println("steps: " + steps_);
    }
    
    public static double latency(){
        return latency_;
    }
    public static void setLoss(double loss){
        loss_ = loss;
    }    
     
    @Override
    public void visPaint(Graphics g)
    {
        _container.visPaint(g);
        _source.paint(g,Color.blue);
        _destination.paint(g,Color.red);
        
        for(Node t : castTargets_)
            t.paint(g, Color.green);
        
        for(Edge edge : treeEdges_)
            edge.paint(g, Color.cyan);
                    
        canvas().repaint();
    }
    
    @Override
    public void processKeyReleased(KeyEvent e)
    {
        super.processKeyReleased(e);         
    }
    
    public static void addTarget(Node n){
        castTargets_.add(n);
    }
    
    public static boolean delivered(){
        
        if(!castTargets_.isEmpty()){
            for(Node n : castTargets_){
                System.out.println(Integer.toString(n.id()));                
            }            
        }
        return castTargets_.isEmpty();
    }
    
    public static double delivery(){        
        return (double)visited_.size() / (visited_.size() + castTargets_.size());
    }
    
    public static int castSize(){
//        for(Node n : castTargets_){
//                System.out.println(Integer.toString(n.id()));                
//        }
        return castTargets_.size();
    }
        
    public static int visited(){
        return visited_.size();
    }
    
    public static int broadcasts(){
        return broadcasts_;
    }
    @Override
    public String toString(){
        return super.toString();
  }
}

