/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package georouting.traversalAlgorithms;

import georouting.Edge;
import georouting.Node;
import georouting.node.ImaginaryNode;
import static georouting.traversalAlgorithms.SteinerNode.getAngle;
import java.awt.Color;
import java.awt.Graphics;
import java.util.ArrayList;

/**
 *
 * @author Jordan
 */
public class StNode {
    private ImaginaryNode   node_;
    private Node            destination_;
    private Edge            parentEdge_;
    private StNode          parent_;
    private ArrayList<StNode>   children_;
    private boolean         virtual_;
    private boolean         active_;
        
    public StNode(float x, float y){
        virtual_ = true;
        node_ = new ImaginaryNode(x, y);
        destination_ = node_;
        children_ = new ArrayList<StNode>();
        active_ = true;
    }
    
    public StNode(Node node){
        virtual_ = false;
        destination_ = node;
        node_ = new ImaginaryNode(node.x(), node.y());
        children_ = new ArrayList<StNode>();
        active_ = true;
    }
    
    public StNode(StNode node){
        virtual_ = node.virtual_;
        children_ = new ArrayList<StNode>();
        node_ = new ImaginaryNode(node.destination_.x(), node.destination_.y());

        if(virtual_){
            destination_ = node_;
        }
        else{
            destination_ = node.destination_;
        }       
    }
        
    public boolean collocated(StNode other){
        return (node_.x() == other.node_.x()
                && node_.y() == other.node_.y());
    }
    
    static public StNode SteinerPoint(StNode sa, StNode sb, StNode sc){
        Node a = sa.destination();
        Node b = sb.destination();
        Node c = sc.destination();
        
        if(sa.collocated(sb))
            return new StNode(sc);
        
        else if(sa.collocated(sc))
            return new StNode(sb);

        else if(sb.collocated(sc))
            return new StNode(sa);
        
        else if(getAcuteAngle(b, a, c) > 2 * Math.PI / 3)
            return new StNode(a);
        
        else if(getAcuteAngle(a, b, c) > 2 * Math.PI / 3)
            return new StNode(b);
        
        else if(getAcuteAngle(a, c, b) > 2 * Math.PI / 3)
            return new StNode(c);
                
        // Compute the Steiner Point of a, b, c
        else{
            // Calculate third points of equillateral triangles
            ImaginaryNode ab = equillateralThird(a, b, c);
            ImaginaryNode ac = equillateralThird(a, c, b);
            
            // Draw line from equillateral third to remaining cross term
            Edge ab_c = ab.imaginaryEdgeWith(c);
            Edge ac_b = ac.imaginaryEdgeWith(b);
            
            // Steiner is located at the intersection of these lines
            ImaginaryNode steiner = ab_c.pointOfIntersection(ac_b);
            
            return new StNode(steiner.x(), steiner.y());
        }        
    }
    
    protected static double getAngle(Node a){
      return Math.atan2(a.y(), a.x());
    }
    
    protected static  double getAngle(Node a, Node b){
      return Math.atan2(a.y() - b.y(), a.x() - b.x());
    }
    
    protected static double getAngle(Node a, Node b, Node c){
        double angle = Math.atan2( c.y() - b.y(), c.x() - b.x())
                 - Math.atan2( a.y() - b.y() , a.x() - b.x() );
        return angle;
    }
    
    protected static double getAcuteAngle(Node a, Node b, Node c){
        double angle = getAngle(a, b, c);
            angle = Math.abs(angle);
            if(angle > Math.PI) angle = Math.PI * 2 - angle;
        return angle;
    }
    
    public double distance(StNode other){
        return node_.distance(other.node_);
    }
    
    private static ImaginaryNode equillateralThird(Node a, Node b, Node c){        
        double rAB = a.distance(b);
        double AB = Math.PI - getAngle(a, b);
        double BAC = getAngle(b, a, c);
        double shift = Math.PI / 3;
        
        //System.out.println(BAC * 180 / Math.PI);
                        
        if(BAC < 0)                 shift *= -1;
        if(Math.abs(BAC) > Math.PI) shift *= -1;        
        
        AB += shift;
                        
        ImaginaryNode ab = new ImaginaryNode(a.x() + rAB * Math.cos(AB), a.y() - rAB * Math.sin(AB));
        
        return ab;
    }
    
    public ArrayList<Node> subDestinations(ArrayList<Node> dests, int t){
        if(!virtual_)
            dests.add(destination_);

//        for(int tabs = 0; tabs < t; tabs++)
//            System.out.print("\t");
//        System.out.println(destination_.id());

        for(StNode child : children_)
            child.subDestinations(dests, t + 1);
        
        return dests;
    }
    
    protected double distanceSum(Node n){
        double distance = virtual_ ? 0.0 : n.distance(destination_);
        
        //System.out.println("d(" + n.id() + ", " + destination_.id() + ") = " + distance);
        
        for(StNode child : children_)
            distance += child.distanceSum(n);
        
        return distance;
    }
    
    protected int subTreeCount(){
        int total = 0;
        if(!virtual_) total++;
        
        for(StNode child : children_)
            total += child.subTreeCount();
        
        return total;
    }
    
    public void addChild(StNode child){
        children_.add(child);
        child.parentEdge_ = node_.imaginaryEdgeWith(child.node_);
        child.parent_ = this;
    }
    
    public ArrayList<StNode> children(){
        return children_;
    }
    
    public Node destination(){
        return destination_;
    }
    
    public Edge parentEdge(){
        return parentEdge_;
    }
    
    public StNode parent(){
        return parent_;
    }
    
    public boolean virtual(){
        return virtual_;
    }
    
    public boolean active(){
        return active_;        
    }
    
    public void deactivate(){
        active_ = false;
    }
    
    public void paint(Graphics g) {
        if(virtual_)
            destination_.paint(g, Color.MAGENTA);
        else
            destination_.paint(g, Color.GREEN);

        // node_.paint(g, Color.MAGENTA);

        if(parentEdge_ != null)
            parentEdge_.paint(g, Color.ORANGE);
        
        for(StNode child : children_)
            child.paint(g);
    }
    
    public StNode getClone(){
        StNode copy = new StNode(this);
        copy.active_ = active_;
        
        for(StNode child : children()){
            StNode copy_kid = child.getClone();
            copy.addChild(copy_kid);
        }
        
        return copy;
    }
    
    public void print(int tier){
        for(int i = 0; i < tier; i++)
            System.out.print("\t");
        System.out.print("{");
        System.out.print("N:<i-");
        System.out.print(node_.id());
        System.out.print(", (");
        System.out.print(node_.x());
        System.out.print(", ");
        System.out.print(node_.y());
        System.out.print(")>, ");
        System.out.print("D:<i-");
        System.out.print(destination_.id());
        System.out.print(", (");
        System.out.print(destination_.x());
        System.out.print(", ");
        System.out.print(destination_.y());
        System.out.print(")> ");
        if(virtual_)
            System.out.print("-V ");
        if(active_)
            System.out.print("-A ");
        System.out.println();
        for(StNode child : children_){
            child.print(tier + 1);
        }
        
    }
}
