/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package georouting.traversalAlgorithms;

import georouting.Edge;
import georouting.Node;
import georouting.node.ImaginaryNode;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.util.ArrayList;

/**
 *
 * @author Jordan
 */
public class SteinerNode {
    ImaginaryNode                   destination;
    public ArrayList<SteinerNode>   children;
    public Edge                     parentEdge;
    public boolean                  active;
    public boolean                  virtual;
    
    public SteinerNode(ImaginaryNode dest){        
        children = new ArrayList<>();
        parentEdge = null;
        active = true;
        destination = dest;
        virtual = false;
    }
    
    public void addChild(SteinerNode child){
        children.add(child);
        child.parentEdge = destination.imaginaryEdgeWith(child.destination);
    }
    
    protected static double getAngle(float ax, float ay){
      return Math.atan2(ay, ax);
    }
    
    protected static  double getAngle(float ax, float ay, float bx, float by){
      return Math.atan2(ay - by, ax - bx);
    }
    
    protected static double getAngle(float ax, float ay, float bx, float by, float cx, float cy){
        double angle = Math.atan2( cy - by, cx - bx)
                 - Math.atan2( ay - by , ax - bx );
        return angle;
    }
    
    protected static double getAcuteAngle(float ax, float ay, float bx, float by, float cx, float cy){
        double angle = getAngle(ax, ay, bx, by, cx, cy);
            angle = Math.abs(angle);
            if(angle > Math.PI) angle = Math.PI * 2 - angle;
        return angle;
    }
    
    
    
    private static ImaginaryNode equillateralThird(float ax, float ay, float bx, float by, float cx, float cy){
        ImaginaryNode a = new ImaginaryNode(ax, ay);
        ImaginaryNode b = new ImaginaryNode(bx, by);
        ImaginaryNode c = new ImaginaryNode(cx, cy);
        
        double rAB = a.distance(b);
        double AB = Math.PI - getAngle(ax, ay, bx, by);
        double BAC = getAngle(bx, by, ax, ay, cx, cy);
        double shift = Math.PI / 3;
        
        //System.out.println(BAC * 180 / Math.PI);
                        
        if(BAC < 0)                 shift *= -1;
        if(Math.abs(BAC) > Math.PI) shift *= -1;        
        
        AB += shift;
                        
        ImaginaryNode ab = new ImaginaryNode(a.x() + rAB * Math.cos(AB), a.y() - rAB * Math.sin(AB));
        
        return ab;
    }
    
    private static Point2D getIntersection(Line2D.Float line1, Line2D.Float line2) {
        float x1,y1, x2,y2, x3,y3, x4,y4;
        x1 = line1.x1; y1 = line1.y1; x2 = line1.x2; y2 = line1.y2;
        x3 = line2.x1; y3 = line2.y1; x4 = line2.x2; y4 = line2.y2;
        float x = ( (x2 - x1)*(x3*y4 - x4*y3) - (x4 - x3)*(x1*y2 - x2*y1) )
                 / ( (x1 - x2)*(y3 - y4) - (y1 - y2)*(x3 - x4) );
        float y = ( (y3 - y4)*(x1*y2 - x2*y1) - (y1 - y2)*(x3*y4 - x4*y3) )
                 / ( (x1 - x2)*(y3 - y4) - (y1 - y2)*(x3 - x4) );

        return new Point2D.Float(x, y);
    }
        
    public static SteinerNode steinerPoint(float ax, float ay, float bx, float by, float cx, float cy){                
        ImaginaryNode steiner;
        boolean isVirtual = false;
        
        if(ax == bx && ay == by)    // a == b, then c
            steiner = new ImaginaryNode(cx, cy);

        else if(ax == cx && ay == cy)
            steiner = new ImaginaryNode(bx, by);
        
        else if(bx == cx && by == cy)
            steiner = new ImaginaryNode(ax, ay);
        
        else if( getAcuteAngle(bx, by, ax, ay, cx, cy) > 2 * Math.PI / 3)
            steiner = new ImaginaryNode(ax, ay);
        
        else if(getAcuteAngle(ax, ay, bx, by, cx, cy) > 2 * Math.PI / 3)
            steiner = new ImaginaryNode(bx, by);
        
        else if(getAcuteAngle(ax, ay, cx, cy, bx, by) > 2 * Math.PI / 3)
            steiner = new ImaginaryNode(cx, cy);
        
        else{
            Node ab = equillateralThird(ax, ay, bx, by, cx, cy);
            Node ac = equillateralThird(ax, ay, cx, cy, bx, by);
            Node bc = equillateralThird(cx, cy, bx, by, ax, ay);        
        
            Line2D.Float ab_c = new Line2D.Float(ax, ay, cx, cy);
            Line2D.Float ac_b = new Line2D.Float(ax, ay, bx, by);
            
            Point2D steinerPt = getIntersection(ab_c, ac_b);
            steiner = new ImaginaryNode(steinerPt.getX(), steinerPt.getY());
        
            //double ATB = getAngle(a, steiner, b) * 180 / Math.PI;
            //double ATC = getAngle(a, steiner, c) * 180 / Math.PI;
            //double BTC = getAngle(b, steiner, c) * 180 / Math.PI;
        
            //System.out.println(ATB + ", " + ATC + ", " + BTC);                               
            isVirtual = true;
        }
        
        SteinerNode st = new SteinerNode(steiner);
        st.virtual = isVirtual;
        return st;
    }    
    
    public double reductionRatio(ImaginaryNode s, ImaginaryNode u, ImaginaryNode v){
        ImaginaryNode t = destination;
        double rr = (1 - (s.distance(t) + t.distance(u) + t.distance(v))
                / (s.distance(u) + s.distance(v)));
        if(rr != rr)
            return 0;        
            
        return rr;
    }
    
    public boolean collocated(SteinerNode other){
        return (destination.x() == other.destination.x()
                && destination.y() == other.destination.y());
    }
    
    public double distance(SteinerNode other){
        return destination.distance(other.destination);
    }
    
    public void paint(Graphics g)
    {
        if(virtual)
            destination.paint(g, Color.PINK);
        else    
            destination.paint(g, Color.green);
        
        if(parentEdge != null)
            parentEdge.paint(g, Color.cyan);
        
        for(SteinerNode child : children)
            child.paint(g);
    }
}
