/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package georouting.traversalAlgorithms;

/**
 *
 * @author Jordan
 */
public class SteinerTuple implements Comparable{
    private StNode s_;
    private StNode t_;
    private StNode u_;
    private StNode v_;    
    
    public SteinerTuple(StNode root, StNode left, StNode right){
        s_ = root;
        u_ = left;
        v_ = right;
        t_ = StNode.SteinerPoint(s_, u_, v_);
    }
    
    public double reductionRatio(){
        return 1 - ((s_.distance(t_) + t_.distance(u_) + t_.distance(v_))
                / (s_.distance(u_) + s_.distance(v_)));
    }
    
    @Override
    public int compareTo(Object other){
        int myRR = (int)(reductionRatio() * 10000000);
        int otherRR = (int)(((SteinerTuple)other).reductionRatio() * 10000000);
        
        return otherRR - myRR;
    }
    
    @Override
    public String toString(){
        // RR
        String str = "RR: " + reductionRatio() + " -> ";
                        
        // s
        str += "{s: " + s_.destination().id();
        if(s_.virtual())
            str += " (v) ";        
        
        str += " - [" + s_.destination().x() + ", " + s_.destination().y() + "]}, ";
        
        // t
        str += "{t: " + t_.destination().id();
        if(t_.virtual())
            str += " (v) ";        
        
        str += " - [" + t_.destination().x() + ", " + t_.destination().y() + "]}, ";
        
        // u
        str += "{u: " + u_.destination().id();
        if(u_.virtual())
            str += " (v) ";        
        
        str += " - [" + u_.destination().x() + ", " + u_.destination().y() + "]}, ";
        
        // v
        str += "{v: " + v_.destination().id();
        if(v_.virtual())
            str += " (v) ";        
        
        str += " - [" + v_.destination().x() + ", " + v_.destination().y() + "]}, ";

        return str;
    }
    
    public StNode u(){
        return u_;       
    }
    
    public StNode v(){
        return v_;       
    }
    
    public StNode t(){
        return t_;       
    }
}
